#pragma once

#ifdef HAVE_UMFPACK

#include <algorithm>
#include <string>

#include <boost/numeric/mtl/operation/two_norm.hpp>
#include <boost/numeric/mtl/interface/umfpack_solve.hpp>

#include <amdis/Output.hpp>
#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>

namespace AMDiS
{
  /**
  * \ingroup Solver
  * \class AMDiS::UmfpackRunner
  * \brief \implements RunnerInterface
  * Wrapper for the external
  *   [UMFPACK solver](http://faculty.cse.tamu.edu/davis/suitesparse.html)
  *
  * This is a direct solver for large sparse matrices.
  */
  template <class Matrix, class Vector>
  class UmfpackRunner;

  template <class T, class Param, class Vector>
  class UmfpackRunner<mtl::compressed2D<T, Param>, Vector>
      : public RunnerInterface<mtl::compressed2D<T, Param>, Vector, Vector>
  {
    using Matrix = mtl::compressed2D<T, Param>;
    using Super = RunnerInterface<Matrix, Vector>;
    using PreconBase = typename Super::PreconBase;

    using SolverType = mtl::mat::umfpack::solver<Matrix>;

  public:
    /// Constructor. Reads UMFPACK parameters from initfile
    UmfpackRunner(std::string const& prefix)
    {
      Parameters::get(prefix + "->store symbolic", storeSymbolic_); // ?
      Parameters::get(prefix + "->symmetric strategy", symmetricStrategy_);
      Parameters::get(prefix + "->alloc init", allocInit_);
    }

    /// Implementation of \ref RunnerInterface::init()
    void init(Matrix const& matrix) override
    {
      try {
        solver_.reset(new SolverType(matrix, symmetricStrategy_, allocInit_));
      } catch (mtl::mat::umfpack::error const& e) {
        error_exit("UMFPACK_ERROR(factorize, {}) = {}", e.code, e.what());
      }
    }

    /// Implementation of \ref RunnerInterface::exit()
    void exit() override {}

    /// Implementation of \ref RunnerBase::solve()
    int solve(Matrix const& A, Vector& x, Vector const& b,
              SolverInfo& solverInfo) override
    {
      AMDIS_FUNCNAME("Umfpack_Runner::solve()");
      test_exit(bool(solver_), "The umfpack solver was not initialized\n");

      int code = 0;
      try {
        code = (*solver_)(x, b);
      } catch (mtl::mat::umfpack::error& e) {
        error_exit("UMFPACK_ERROR(solve, {}) = {}", e.code, e.what());
      }

      auto r = Vector(b);
      if (two_norm(x) != 0)
        r -= A * x;

      solverInfo.setAbsResidual(two_norm(r));
      solverInfo.setError(code);

      return code;
    }

  protected:
    std::shared_ptr<SolverType> solver_;

    int storeSymbolic_ = 0;
    int symmetricStrategy_ = 0;
    double allocInit_ = 0.7;
  };
}

#endif // HAVE_UMFPACK
